@@ -22,6 +22,10 @@
 #include "xslt.h"
 #include "xsltInternals.h"
 #include "xsltutils.h"
+#include "imports.h"
+#include "templates.h"
+#include "keys.h"
+#include "pattern.h"
 
 /* #define DEBUG_PARSING */
 
@@ -55,13 +59,16 @@ struct _xsltStepOp {
     xsltOp op;
     xmlChar *value;
     xmlChar *value2;
+    xmlChar *value3;
 };
 
 typedef struct _xsltCompMatch xsltCompMatch;
 typedef xsltCompMatch *xsltCompMatchPtr;
 struct _xsltCompMatch {
     struct _xsltCompMatch *next; /* siblings in the name hash */
-    float priority;                /* the priority */
+    float priority;              /* the priority */
+    const xmlChar *mode;         /* the mode */
+    const xmlChar *modeURI;      /* the mode URI */
     xsltTemplatePtr template;    /* the associated template */
 
     /* TODO fix the statically allocated size steps[] */
@@ -87,13 +94,15 @@ struct _xsltParserContext {
 
 /**
  * xsltNewCompMatch:
+ * @mode:  the mode name or NULL
+ * @modeURI:  the mode URI or NULL
  *
  * Create a new XSLT CompMatch
  *
  * Returns the newly allocated xsltCompMatchPtr or NULL in case of error
  */
 xsltCompMatchPtr
-xsltNewCompMatch(void) {
+xsltNewCompMatch(const xmlChar *mode, const xmlChar *modeURI) {
     xsltCompMatchPtr cur;
 
     cur = (xsltCompMatchPtr) xmlMalloc(sizeof(xsltCompMatch));
@@ -104,6 +113,8 @@ xsltNewCompMatch(void) {
     }
     memset(cur, 0, sizeof(xsltCompMatch));
     cur->maxStep = 20;
+    cur->mode = xmlStrdup(mode);
+    cur->modeURI = xmlStrdup(modeURI);
     return(cur);
 }
 
@@ -120,12 +131,18 @@ xsltFreeCompMatch(xsltCompMatchPtr comp) {
 
     if (comp == NULL)
 	return;
+    if (comp->mode != NULL)
+	xmlFree((xmlChar *)comp->mode);
+    if (comp->modeURI != NULL)
+	xmlFree((xmlChar *)comp->modeURI);
     for (i = 0;i < comp->nbStep;i++) {
 	op = &comp->steps[i];
 	if (op->value != NULL)
 	    xmlFree(op->value);
 	if (op->value2 != NULL)
 	    xmlFree(op->value2);
+	if (op->value3 != NULL)
+	    xmlFree(op->value3);
     }
     memset(comp, -1, sizeof(xsltCompMatch));
     xmlFree(comp);
@@ -209,6 +226,33 @@ xsltCompMatchAdd(xsltCompMatchPtr comp, xsltOp op, xmlChar *value,
     return(0);
 }
 
+/**
+ * xsltSwapTopCompMatch:
+ * @comp:  the compiled match expression
+ *
+ * reverse the two top steps.
+ */
+void
+xsltSwapTopCompMatch(xsltCompMatchPtr comp) {
+    int i;
+    int j = comp->nbStep - 1;
+
+    if (j > 0) {
+	register xmlChar *tmp;
+	register xsltOp op;
+	i = j - 1;
+	tmp = comp->steps[i].value;
+	comp->steps[i].value = comp->steps[j].value;
+	comp->steps[j].value = tmp;
+	tmp = comp->steps[i].value2;
+	comp->steps[i].value2 = comp->steps[j].value2;
+	comp->steps[j].value2 = tmp;
+	op = comp->steps[i].op;
+	comp->steps[i].op = comp->steps[j].op;
+	comp->steps[j].op = op;
+    }
+}
+
 /**
  * xsltReverseCompMatch:
  * @comp:  the compiled match expression
@@ -246,25 +290,51 @@ xsltReverseCompMatch(xsltCompMatchPtr comp) {
 
 /**
  * xsltTestCompMatch:
+ * @ctxt:  a XSLT process context
  * @comp: the precompiled pattern
  * @node: a node
+ * @mode:  the mode name or NULL
+ * @modeURI:  the mode URI or NULL
  *
  * Test wether the node matches the pattern
  *
  * Returns 1 if it matches, 0 if it doesn't and -1 in case of failure
  */
 int
-xsltTestCompMatch(xsltCompMatchPtr comp, xmlNodePtr node) {
+xsltTestCompMatch(xsltTransformContextPtr ctxt, xsltCompMatchPtr comp,
+	          xmlNodePtr node, const xmlChar *mode,
+		  const xmlChar *modeURI) {
     int i;
-    xsltStepOpPtr step;
+    xsltStepOpPtr step, select = NULL;
 
-    if ((comp == NULL) || (node == NULL)) {
+    if ((comp == NULL) || (node == NULL) || (ctxt == NULL)) {
         xsltGenericError(xsltGenericErrorContext,
 		"xsltTestCompMatch: null arg\n");
         return(-1);
     }
+    if (mode != NULL) {
+	if (comp->mode == NULL)
+	    return(0);
+	if ((comp->mode != mode) && (!xmlStrEqual(comp->mode, mode)))
+	    return(0);
+    } else {
+	if (comp->mode != NULL)
+	    return(0);
+    }
+    if (modeURI != NULL) {
+	if (comp->modeURI == NULL)
+	    return(0);
+	if ((comp->modeURI != modeURI) &&
+	    (!xmlStrEqual(comp->modeURI, modeURI)))
+	    return(0);
+    } else {
+	if (comp->modeURI != NULL)
+	    return(0);
+    }
     for (i = 0;i < comp->nbStep;i++) {
 	step = &comp->steps[i];
+	if (step->op != XSLT_OP_PREDICATE)
+	    select = step;
 	switch (step->op) {
             case XSLT_OP_END:
 		return(1);
@@ -362,16 +432,35 @@ xsltTestCompMatch(xsltCompMatchPtr comp, xmlNodePtr node) {
 				break;
 			}
 		    }
+		    node = node->parent;
 		}
 		if (node == NULL)
 		    return(0);
 		continue;
-            case XSLT_OP_ID:
-		TODO /* Handle IDs, might be done differently */
+            case XSLT_OP_ID: {
+		/* TODO Handle IDs decently, must be done differently */
+		xmlAttrPtr id;
+
+		id = xmlGetID(node->doc, step->value);
+		if ((id == NULL) || (id->parent != node))
+		    return(0);
 		break;
-            case XSLT_OP_KEY:
-		TODO /* Handle Keys, might be done differently */
+	    }
+            case XSLT_OP_KEY: {
+		xmlNodeSetPtr list;
+		int i;
+
+		list = xsltGetKey(ctxt, step->value,
+			          step->value3, step->value2);
+		if (list == NULL)
+		    return(0);
+		for (i = 0;i < list->nodeNr;i++)
+		    if (list->nodeTab[i] == node)
+			break;
+		if (i >= list->nodeNr)
+		    return(0);
 		break;
+	    }
             case XSLT_OP_NS:
 		/* Namespace test */
 		if (node->ns == NULL) {
@@ -394,13 +483,65 @@ xsltTestCompMatch(xsltCompMatchPtr comp, xmlNodePtr node) {
 			return(0);
 		}
 		break;
-	    case XSLT_OP_PREDICATE:
-		TODO /* Handle Predicate */
+	    case XSLT_OP_PREDICATE: {
+		xmlNodePtr oldNode;
+		int oldCS, oldCP;
+		int pos = 0, len = 0;
+		/*
+		 * Depending on the last selection, one may need to
+		 * recompute contextSize and proximityPosition.
+		 */
+		oldCS = ctxt->xpathCtxt->contextSize;
+		oldCP = ctxt->xpathCtxt->proximityPosition;
+		if ((select != NULL) &&
+		    (select->op == XSLT_OP_ELEM) &&
+		    (select->value != NULL) &&
+		    (node->type == XML_ELEMENT_NODE) &&
+		    (node->parent != NULL)) {
+
+		    /* TODO: cache those informations ?!? */
+		    xmlNodePtr siblings = node->parent->children;
+
+		    while (siblings != NULL) {
+			if (siblings->type == XML_ELEMENT_NODE) {
+			    if (siblings == node) {
+				len++;
+				pos = len;
+			    } else if (xmlStrEqual(node->name,
+				       siblings->name)) {
+				len++;
+			    }
+			}
+			siblings = siblings->next;
+		    }
+		    if (pos != 0) {
+			ctxt->xpathCtxt->contextSize = len;
+			ctxt->xpathCtxt->proximityPosition = pos;
+		    }
+		}
+		oldNode = ctxt->node;
+		ctxt->node = node;
+
+		if ((step->value == NULL) ||
+		    (!xsltEvalXPathPredicate(ctxt, step->value))) {
+		    if (pos != 0) {
+			ctxt->xpathCtxt->contextSize = oldCS;
+			ctxt->xpathCtxt->proximityPosition = oldCP;
+		    }
+		    ctxt->node = oldNode;
+		    return(0);
+		}
+		if (pos != 0) {
+		    ctxt->xpathCtxt->contextSize = oldCS;
+		    ctxt->xpathCtxt->proximityPosition = oldCP;
+		}
+		ctxt->node = oldNode;
 		break;
+	    }
             case XSLT_OP_PI:
 		if (node->type != XML_PI_NODE)
 		    return(0);
-		if (step->value == NULL) {
+		if (step->value != NULL) {
 		    if (!xmlStrEqual(step->value, node->name))
 			return(0);
 		}
@@ -455,6 +596,9 @@ xsltTestCompMatch(xsltCompMatchPtr comp, xmlNodePtr node) {
 #define PUSH(op, val, val2) 						\
     if (xsltCompMatchAdd(ctxt->comp, (op), (val), (val2))) goto error;
 
+#define SWAP() 						\
+    xsltSwapTopCompMatch(ctxt->comp);
+
 #define XSLT_ERROR(X)							\
     { xsltError(ctxt, __FILE__, __LINE__, X);			\
       ctxt->error = (X); return; }
@@ -633,7 +777,8 @@ xsltCompileIdKeyPattern(xsltParserContextPtr ctxt, xmlChar *name, int aid) {
 	    return;
 	}
 	NEXT;
-	PUSH(XSLT_OP_KEY, lit, NULL);
+	/* TODO: support namespace in keys */
+	PUSH(XSLT_OP_KEY, lit, lit2);
     } else if (xmlStrEqual(name, (const xmlChar *)"processing-instruction")) {
 	NEXT;
 	SKIP_BLANKS;
@@ -696,10 +841,8 @@ xsltCompileIdKeyPattern(xsltParserContextPtr ctxt, xmlChar *name, int aid) {
 	return;
     }
 error:
-    if (lit != NULL)
-	xmlFree(lit);
-    if (lit2 != NULL)
-	xmlFree(lit2);
+    if (name != NULL)
+	xmlFree(name);
 }
 
 /**
@@ -729,6 +872,12 @@ xsltCompileStepPattern(xsltParserContextPtr ctxt, xmlChar *token) {
 
     SKIP_BLANKS;
     if ((token == NULL) && (CUR == '@')) {
+	NEXT;
+	if (CUR == '*') {
+	    NEXT;
+	    PUSH(XSLT_OP_ATTR, NULL, NULL);
+	    return;
+	}
 	token = xsltScanName(ctxt);
 	if (token == NULL) {
 	    xsltGenericError(xsltGenericErrorContext,
@@ -818,9 +967,11 @@ parse_predicate:
 	    ctxt->error = 1;
 	    goto error;
         }
-	NEXT;
 	ret = xmlStrndup(q, CUR_PTR - q);
 	PUSH(XSLT_OP_PREDICATE, ret, NULL);
+	/* push the predicate lower than local test */
+	SWAP();
+	NEXT;
     }
     return;
 error:
@@ -908,12 +1059,14 @@ xsltCompileLocationPathPattern(xsltParserContextPtr ctxt) {
 	}
     } else if (CUR == '*') {
 	xsltCompileRelativePathPattern(ctxt, NULL);
+    } else if (CUR == '@') {
+	xsltCompileRelativePathPattern(ctxt, NULL);
     } else {
 	xmlChar *name;
 	name = xsltScanName(ctxt);
 	if (name == NULL) {
 	    xsltGenericError(xsltGenericErrorContext,
-		    "xsltCompilePattern : Name expected\n");
+		    "xsltCompileLocationPathPattern : Name expected\n");
 	    ctxt->error = 1;
 	    return;
 	}
@@ -943,6 +1096,8 @@ error:
 /**
  * xsltCompilePattern:
  * @pattern an XSLT pattern
+ * @mode:  the mode name or NULL
+ * @modeURI:  the mode URI or NULL
  *
  * Compile the XSLT pattern and generates a precompiled form suitable
  * for fast matching.
@@ -955,7 +1110,8 @@ error:
  */
 
 xsltCompMatchPtr
-xsltCompilePattern(const xmlChar *pattern) {
+xsltCompilePattern(const xmlChar *pattern, const xmlChar *mode,
+	           const xmlChar *modeURI) {
     xsltParserContextPtr ctxt;
     xsltCompMatchPtr ret;
     const xmlChar *cur;
@@ -981,7 +1137,7 @@ xsltCompilePattern(const xmlChar *pattern) {
     ctxt = xsltNewParserContext();
     if (ctxt == NULL)
 	return(NULL);
-    ret = xsltNewCompMatch();
+    ret = xsltNewCompMatch(mode, modeURI);
     if (ret == NULL) {
 	xsltFreeParserContext(ctxt);
 	return(NULL);
@@ -1046,14 +1202,17 @@ error:
  * xsltAddTemplate:
  * @style: an XSLT stylesheet
  * @cur: an XSLT template
+ * @mode:  the mode name or NULL
+ * @modeURI:  the mode URI or NULL
  *
  * Register the XSLT pattern associated to @cur
  *
  * Returns -1 in case of error, 0 otherwise
  */
 int
-xsltAddTemplate(xsltStylesheetPtr style, xsltTemplatePtr cur) {
-    xsltCompMatchPtr pat, list, *top;
+xsltAddTemplate(xsltStylesheetPtr style, xsltTemplatePtr cur,
+	        const xmlChar *mode, const xmlChar *modeURI) {
+    xsltCompMatchPtr pat, list, *top = NULL;
     const xmlChar *name = NULL;
     xmlChar *p, *pattern, tmp;
 
@@ -1078,7 +1237,7 @@ next_pattern:
 
     tmp = *p;
     *p = 0;
-    pat = xsltCompilePattern(pattern);
+    pat = xsltCompilePattern(pattern, mode, modeURI);
     *p = tmp;
     if (tmp != 0)
 	p++;
@@ -1102,14 +1261,17 @@ next_pattern:
         case XSLT_OP_CHILD:
         case XSLT_OP_PARENT:
         case XSLT_OP_ANCESTOR:
-        case XSLT_OP_ID:
-        case XSLT_OP_KEY:
         case XSLT_OP_NS:
              name = pat->steps[0].value;
 	     break;
         case XSLT_OP_ROOT:
              top = (xsltCompMatchPtr *) &(style->rootMatch);
 	     break;
+        case XSLT_OP_KEY:
+             top = (xsltCompMatchPtr *) &(style->keyMatch);
+	     break;
+        case XSLT_OP_ID:
+	     /* TODO optimize ID !!! */
         case XSLT_OP_ALL:
              top = (xsltCompMatchPtr *) &(style->elemMatch);
 	     break;
@@ -1145,7 +1307,7 @@ next_pattern:
     }
     if (name != NULL) {
 	if (style->templatesHash == NULL) {
-	    style->templatesHash = xmlHashCreate(0);
+	    style->templatesHash = xmlHashCreate(1024);
 	    if (style->templatesHash == NULL) {
 		xsltFreeCompMatch(pat);
 		return(-1);
@@ -1154,15 +1316,17 @@ next_pattern:
 	    xsltGenericDebug(xsltGenericDebugContext,
 		    "xsltAddTemplate: created template hash\n");
 #endif
-	    xmlHashAddEntry(style->templatesHash, name, pat);
+	    xmlHashAddEntry3(style->templatesHash, name, mode, modeURI, pat);
 #ifdef DEBUG_PARSING
 	    xsltGenericDebug(xsltGenericDebugContext,
 		    "xsltAddTemplate: added new hash %s\n", name);
 #endif
 	} else {
-	    list = (xsltCompMatchPtr) xmlHashLookup(style->templatesHash, name);
+	    list = (xsltCompMatchPtr) xmlHashLookup3(style->templatesHash,
+		                      name, mode, modeURI);
 	    if (list == NULL) {
-		xmlHashAddEntry(style->templatesHash, name, pat);
+		xmlHashAddEntry3(style->templatesHash, name,
+			         mode, modeURI, pat);
 #ifdef DEBUG_PARSING
 		xsltGenericDebug(xsltGenericDebugContext,
 			"xsltAddTemplate: added new hash %s\n", name);
@@ -1175,7 +1339,8 @@ next_pattern:
 		 */
 		if (list->priority <= pat->priority) {
 		    pat->next = list;
-		    xmlHashUpdateEntry(style->templatesHash, name, pat, NULL);
+		    xmlHashUpdateEntry3(style->templatesHash, name,
+			                mode, modeURI, pat, NULL);
 #ifdef DEBUG_PARSING
 		    xsltGenericDebug(xsltGenericDebugContext,
 			    "xsltAddTemplate: added head hash for %s\n", name);
@@ -1184,6 +1349,7 @@ next_pattern:
 		    while (list->next != NULL) {
 			if (list->next->priority <= pat->priority)
 			    break;
+			list = list->next;
 		    }
 		    pat->next = list->next;
 		    list->next = pat;
@@ -1202,6 +1368,7 @@ next_pattern:
 	    while (list->next != NULL) {
 		if (list->next->priority <= pat->priority)
 		    break;
+		list = list->next;
 	    }
 	    pat->next = list->next;
 	    list->next = pat;
@@ -1219,38 +1386,100 @@ next_pattern:
 
 /**
  * xsltGetTemplate:
- * @style: an XSLT stylesheet
- * @node: an XML Node
+ * @ctxt:  a XSLT process context
+ * @mode:  the mode name or NULL
  *
  * Finds the template applying to this node
  *
  * Returns the xsltTemplatePtr or NULL if not found
  */
 xsltTemplatePtr
-xsltGetTemplate(xsltStylesheetPtr style, xmlNodePtr node) {
+xsltGetTemplate(xsltTransformContextPtr ctxt, xmlNodePtr node) {
+    xsltStylesheetPtr style;
     xsltTemplatePtr ret = NULL;
     const xmlChar *name = NULL;
     xsltCompMatchPtr list = NULL;
 
-    if ((style == NULL) || (node == NULL))
+    if ((ctxt == NULL) || (node == NULL))
 	return(NULL);
 
-    /* TODO : handle IDs/keys here ! */
-    if (style->templatesHash != NULL) {
+    style = ctxt->style;
+    while (style != NULL) {
+	/* TODO : handle IDs/keys here ! */
+	if (style->templatesHash != NULL) {
+	    /*
+	     * Use the top name as selector
+	     */
+	    switch (node->type) {
+		case XML_ELEMENT_NODE:
+		case XML_ATTRIBUTE_NODE:
+		case XML_PI_NODE:
+		    name = node->name;
+		    break;
+		case XML_DOCUMENT_NODE:
+		case XML_HTML_DOCUMENT_NODE:
+		case XML_TEXT_NODE:
+		case XML_CDATA_SECTION_NODE:
+		case XML_COMMENT_NODE:
+		case XML_ENTITY_REF_NODE:
+		case XML_ENTITY_NODE:
+		case XML_DOCUMENT_TYPE_NODE:
+		case XML_DOCUMENT_FRAG_NODE:
+		case XML_NOTATION_NODE:
+		case XML_DTD_NODE:
+		case XML_ELEMENT_DECL:
+		case XML_ATTRIBUTE_DECL:
+		case XML_ENTITY_DECL:
+		case XML_NAMESPACE_DECL:
+		case XML_XINCLUDE_START:
+		case XML_XINCLUDE_END:
+		    break;
+		default:
+		    return(NULL);
+
+	    }
+	}
+	if (name != NULL) {
+	    /*
+	     * find the list of appliable expressions based on the name
+	     */
+	    list = (xsltCompMatchPtr) xmlHashLookup3(style->templatesHash,
+					     name, ctxt->mode, ctxt->modeURI);
+	}
+	while (list != NULL) {
+	    if (xsltTestCompMatch(ctxt, list, node,
+			          ctxt->mode, ctxt->modeURI)) {
+		ret = list->template;
+		break;
+	    }
+	    list = list->next;
+	}
+	list = NULL;
+
 	/*
-	 * Use the top name as selector
+	 * find alternate generic matches
 	 */
 	switch (node->type) {
 	    case XML_ELEMENT_NODE:
+		list = style->elemMatch;
+		break;
 	    case XML_ATTRIBUTE_NODE:
+		list = style->attrMatch;
+		break;
 	    case XML_PI_NODE:
-		name = node->name;
+		list = style->piMatch;
 		break;
 	    case XML_DOCUMENT_NODE:
 	    case XML_HTML_DOCUMENT_NODE:
+		list = style->rootMatch;
+		break;
 	    case XML_TEXT_NODE:
 	    case XML_CDATA_SECTION_NODE:
+		list = style->textMatch;
+		break;
 	    case XML_COMMENT_NODE:
+		list = style->commentMatch;
+		break;
 	    case XML_ENTITY_REF_NODE:
 	    case XML_ENTITY_NODE:
 	    case XML_DOCUMENT_TYPE_NODE:
@@ -1263,76 +1492,42 @@ xsltGetTemplate(xsltStylesheetPtr style, xmlNodePtr node) {
 	    case XML_NAMESPACE_DECL:
 	    case XML_XINCLUDE_START:
 	    case XML_XINCLUDE_END:
-		return(NULL);
+		break;
 	    default:
-		return(NULL);
+		break;
 
 	}
-    }
-    if (name != NULL) {
+	while ((list != NULL) &&
+	       ((ret == NULL)  || (list->priority > ret->priority))) {
+	    if (xsltTestCompMatch(ctxt, list, node,
+			          ctxt->mode, ctxt->modeURI)) {
+		ret = list->template;
+		break;
+	    }
+	    list = list->next;
+	}
+	if (node->_private != NULL) {
+	    list = style->keyMatch;
+	    while ((list != NULL) &&
+		   ((ret == NULL)  || (list->priority > ret->priority))) {
+		if (xsltTestCompMatch(ctxt, list, node,
+				      ctxt->mode, ctxt->modeURI)) {
+		    ret = list->template;
+		    break;
+		}
+		list = list->next;
+	    }
+	}
+
+	if (ret != NULL)
+	    return(ret);
+
 	/*
-	 * find the list of appliable expressions based on the name
+	 * Cycle on next stylesheet import.
 	 */
-	list = (xsltCompMatchPtr) xmlHashLookup(style->templatesHash, name);
+	style = xsltNextImport(style);
     }
-    while (list != NULL) {
-	if (xsltTestCompMatch(list, node)) {
-	    ret = list->template;
-	    break;
-	}
-	list = list->next;
-    }
-    list = NULL;
-
-    /*
-     * find alternate generic matches
-     */
-    switch (node->type) {
-        case XML_ELEMENT_NODE:
-	    list = style->elemMatch;
-	    break;
-        case XML_ATTRIBUTE_NODE:
-	    list = style->attrMatch;
-	    break;
-        case XML_PI_NODE:
-	    list = style->piMatch;
-	    break;
-        case XML_DOCUMENT_NODE:
-        case XML_HTML_DOCUMENT_NODE:
-	    list = style->rootMatch;
-	    break;
-        case XML_TEXT_NODE:
-        case XML_CDATA_SECTION_NODE:
-	    list = style->textMatch;
-	    break;
-        case XML_COMMENT_NODE:
-	    list = style->commentMatch;
-	    break;
-        case XML_ENTITY_REF_NODE:
-        case XML_ENTITY_NODE:
-        case XML_DOCUMENT_TYPE_NODE:
-        case XML_DOCUMENT_FRAG_NODE:
-        case XML_NOTATION_NODE:
-        case XML_DTD_NODE:
-        case XML_ELEMENT_DECL:
-        case XML_ATTRIBUTE_DECL:
-        case XML_ENTITY_DECL:
-        case XML_NAMESPACE_DECL:
-        case XML_XINCLUDE_START:
-        case XML_XINCLUDE_END:
-	    break;
-	default:
-	    break;
-
-    }
-    while ((list != NULL) &&
-	   ((ret == NULL)  || (list->priority > ret->priority))) {
-	if (xsltTestCompMatch(list, node)) {
-	    ret = list->template;
-	    break;
-	}
-    }
-    return(ret);
+    return(NULL);
 }
 
 
@@ -1349,6 +1544,8 @@ xsltFreeTemplateHashes(xsltStylesheetPtr style) {
 		    (xmlHashDeallocator) xsltFreeCompMatchList);
     if (style->rootMatch != NULL)
         xsltFreeCompMatchList(style->rootMatch);
+    if (style->keyMatch != NULL)
+        xsltFreeCompMatchList(style->keyMatch);
     if (style->elemMatch != NULL)
         xsltFreeCompMatchList(style->elemMatch);
     if (style->attrMatch != NULL)
@@ -1363,36 +1560,3 @@ xsltFreeTemplateHashes(xsltStylesheetPtr style) {
         xsltFreeCompMatchList(style->commentMatch);
 }
 
-/**
- * xsltFindTemplate:
- * @style: an XSLT stylesheet
- * @name: the template name
- * @nameURI: the template name URI
- *
- * Finds the named template.
- *
- * Returns the xsltTemplatePtr or NULL if not found
- */
-xsltTemplatePtr
-xsltFindTemplate(xsltStylesheetPtr style, const xmlChar *name,
-	        const xmlChar *nameURI) {
-    xsltTemplatePtr cur;
-
-    if ((style == NULL) || (name == NULL))
-	return(NULL);
-
-    /* TODO: apply stylesheet import order */
-    cur = style->templates;
-    while (cur != NULL) {
-	if (xmlStrEqual(name, cur->name)) {
-	    if (((nameURI == NULL) && (cur->nameURI == NULL)) ||
-		((nameURI != NULL) && (cur->nameURI != NULL) &&
-		 (xmlStrEqual(nameURI, cur->nameURI)))) {
-		return(cur);
-	    }
-	}
-	cur = cur->next;
-    }
-    return(NULL);
-}
-
