@@ -7,7 +7,7 @@
  */
 
 #ifndef __XML_XSLT_PATTERN_H__
-#define __XML_XSLT_H__
+#define __XML_XSLT_PATTERN_H__
 
 #include "xsltInternals.h"
 
@@ -16,16 +16,15 @@ extern "C" {
 #endif
 
 int		xsltAddTemplate		(xsltStylesheetPtr style,
-					 xsltTemplatePtr cur);
-xsltTemplatePtr	xsltGetTemplate		(xsltStylesheetPtr style,
+					 xsltTemplatePtr cur,
+					 const xmlChar *mode,
+					 const xmlChar *modeURI);
+xsltTemplatePtr	xsltGetTemplate		(xsltTransformContextPtr ctxt,
 					 xmlNodePtr node);
 void		xsltFreeTemplateHashes	(xsltStylesheetPtr style);
-xsltTemplatePtr xsltFindTemplate	(xsltStylesheetPtr style,
-					 const xmlChar *name,
-					 const xmlChar *nameURI);
 #ifdef __cplusplus
 }
 #endif
 
-#endif /* __XML_XSLT_H__ */
+#endif /* __XML_XSLT_PATTERN_H__ */
 
