@@ -18,15 +18,21 @@
 #include <libxml/tree.h>
 #include <libxml/valid.h>
 #include <libxml/hash.h>
+#include <libxml/uri.h>
 #include <libxml/xmlerror.h>
 #include <libxml/parserInternals.h>
 #include "xslt.h"
 #include "xsltInternals.h"
 #include "pattern.h"
 #include "variables.h"
+#include "namespaces.h"
+#include "attributes.h"
 #include "xsltutils.h"
+#include "imports.h"
+#include "keys.h"
 
 #define DEBUG_PARSING
+/* #define DEBUG_BLANKS */
 
 /*
  * Useful macros
@@ -67,6 +73,102 @@ xsltIsBlank(xmlChar *str) {
  *		Routines to handle XSLT data structures			*
  *									*
  ************************************************************************/
+static xsltDecimalFormatPtr
+xsltNewDecimalFormat(xmlChar *name)
+{
+    xsltDecimalFormatPtr self;
+
+    self = xmlMalloc(sizeof(xsltDecimalFormat));
+    if (self != NULL) {
+	self->next = NULL;
+	self->name = name;
+	
+	/* Default values */
+	self->digit = xmlStrdup(BAD_CAST("0"));
+	self->patternSeparator = xmlStrdup(BAD_CAST(";"));
+	self->decimalPoint = xmlStrdup(BAD_CAST("."));
+	self->grouping = xmlStrdup(BAD_CAST(","));
+	self->percent = xmlStrdup(BAD_CAST("%"));
+	self->permille = xmlStrdup(BAD_CAST("?"));
+	self->zeroDigit = xmlStrdup(BAD_CAST("#"));
+	self->minusSign = xmlStrdup(BAD_CAST("-"));
+	self->infinity = xmlStrdup(BAD_CAST("Infinity"));
+	self->noNumber = xmlStrdup(BAD_CAST("NaN"));
+    }
+    return self;
+}
+
+static void
+xsltFreeDecimalFormat(xsltDecimalFormatPtr self)
+{
+    if (self != NULL) {
+	if (self->digit)
+	    xmlFree(self->digit);
+	if (self->patternSeparator)
+	    xmlFree(self->patternSeparator);
+	if (self->decimalPoint)
+	    xmlFree(self->decimalPoint);
+	if (self->grouping)
+	    xmlFree(self->grouping);
+	if (self->percent)
+	    xmlFree(self->percent);
+	if (self->permille)
+	    xmlFree(self->permille);
+	if (self->zeroDigit)
+	    xmlFree(self->zeroDigit);
+	if (self->minusSign)
+	    xmlFree(self->minusSign);
+	if (self->infinity)
+	    xmlFree(self->infinity);
+	if (self->noNumber)
+	    xmlFree(self->noNumber);
+	if (self->name)
+	    xmlFree(self->name);
+	xmlFree(self);
+    }
+}
+
+static void
+xsltFreeDecimalFormatList(xsltStylesheetPtr self)
+{
+    xsltDecimalFormatPtr iter;
+    xsltDecimalFormatPtr tmp;
+
+    if (self == NULL)
+	return;
+    
+    iter = self->decimalFormat;
+    while (iter != NULL) {
+	tmp = iter->next;
+	xsltFreeDecimalFormat(iter);
+	iter = tmp;
+    }
+}
+
+/**
+ * xsltDecimalFormatGetByName:
+ * @sheet: the XSLT stylesheet
+ * @name: the decimal-format name to find
+ *
+ * Find decimal-format by name
+ */
+xsltDecimalFormatPtr
+xsltDecimalFormatGetByName(xsltStylesheetPtr sheet, xmlChar *name)
+{
+    xsltDecimalFormatPtr result;
+
+    if (name == NULL)
+	return sheet->decimalFormat;
+    
+    for (result = sheet->decimalFormat->next;
+	 result != NULL;
+	 result = result->next) {
+	if (xmlStrEqual(name, result->name))
+	    break; /* for */
+    }
+    return result;
+}
+
 
 /**
  * xsltNewTemplate:
@@ -146,10 +248,28 @@ xsltNewStylesheet(void) {
     memset(cur, 0, sizeof(xsltStylesheet));
     cur->omitXmlDeclaration = -1;
     cur->standalone = -1;
+    cur->decimalFormat = xsltNewDecimalFormat(NULL);
     cur->indent = -1;
     return(cur);
 }
 
+/**
+ * xsltFreeStylesheetList:
+ * @sheet:  an XSLT stylesheet list
+ *
+ * Free up the memory allocated by the list @sheet
+ */
+void
+xsltFreeStylesheetList(xsltStylesheetPtr sheet) {
+    xsltStylesheetPtr next;
+
+    while (sheet != NULL) {
+	next = sheet->next;
+	xsltFreeStylesheet(sheet);
+	sheet = next;
+    }
+}
+
 /**
  * xsltFreeStylesheet:
  * @sheet:  an XSLT stylesheet
@@ -161,8 +281,12 @@ xsltFreeStylesheet(xsltStylesheetPtr sheet) {
     if (sheet == NULL)
 	return;
 
+    xsltFreeKeys(sheet);
     xsltFreeTemplateHashes(sheet);
+    xsltFreeDecimalFormatList(sheet);
     xsltFreeTemplateList(sheet->templates);
+    xsltFreeAttributeSetsHashes(sheet);
+    xsltFreeNamespaceAliasHashes(sheet);
     if (sheet->doc != NULL)
 	xmlFreeDoc(sheet->doc);
     if (sheet->variables != NULL)
@@ -178,6 +302,9 @@ xsltFreeStylesheet(xsltStylesheetPtr sheet) {
     if (sheet->doctypeSystem != NULL) xmlFree(sheet->doctypeSystem);
     if (sheet->mediaType != NULL) xmlFree(sheet->mediaType);
 
+    if (sheet->imports != NULL)
+	xsltFreeStylesheetList(sheet->imports);
+
     memset(sheet, -1, sizeof(xsltStylesheet));
     xmlFree(sheet);
 }
@@ -345,6 +472,108 @@ xsltParseStylesheetOutput(xsltStylesheetPtr style, xmlNodePtr cur) {
     }
 }
 
+/**
+ * xsltParseStylesheetDecimalFormat:
+ * @style:  the XSLT stylesheet
+ * @cur:  the "decimal-format" element
+ *
+ * parse an XSLT stylesheet decimal-format element and
+ * and record the formatting characteristics
+ */
+void
+xsltParseStylesheetDecimalFormat(xsltStylesheetPtr sheet, xmlNodePtr cur)
+{
+    xmlChar *prop;
+    xsltDecimalFormatPtr format;
+    xsltDecimalFormatPtr iter;
+    
+    if ((cur == NULL) || (sheet == NULL))
+	return;
+
+    format = sheet->decimalFormat;
+    
+    prop = xmlGetNsProp(cur, BAD_CAST("name"), XSLT_NAMESPACE);
+    if (prop != NULL) {
+	format = xsltDecimalFormatGetByName(sheet, prop);
+	if (format != NULL) {
+	    xsltGenericError(xsltGenericErrorContext,
+			     "xsltParseStylesheetDecimalFormat: %s already exists\n", prop);
+	    return;
+	}
+	format = xsltNewDecimalFormat(prop);
+	if (format == NULL) {
+	    xsltGenericError(xsltGenericErrorContext,
+			     "xsltParseStylesheetDecimalFormat: failed creating new decimal-format\n");
+	    return;
+	}
+	/* Append new decimal-format structure */
+	for (iter = sheet->decimalFormat; iter->next; iter = iter->next)
+	    ;
+	if (iter)
+	    iter->next = format;
+    }
+
+    prop = xmlGetNsProp(cur, (const xmlChar *)"decimal-separator", XSLT_NAMESPACE);
+    if (prop != NULL) {
+	if (format->decimalPoint != NULL) xmlFree(format->decimalPoint);
+	format->decimalPoint  = prop;
+    }
+    
+    prop = xmlGetNsProp(cur, (const xmlChar *)"grouping-separator", XSLT_NAMESPACE);
+    if (prop != NULL) {
+	if (format->grouping != NULL) xmlFree(format->grouping);
+	format->grouping  = prop;
+    }
+
+    prop = xmlGetNsProp(cur, (const xmlChar *)"infinity", XSLT_NAMESPACE);
+    if (prop != NULL) {
+	if (format->infinity != NULL) xmlFree(format->infinity);
+	format->infinity  = prop;
+    }
+    
+    prop = xmlGetNsProp(cur, (const xmlChar *)"minus-sign", XSLT_NAMESPACE);
+    if (prop != NULL) {
+	if (format->minusSign != NULL) xmlFree(format->minusSign);
+	format->minusSign  = prop;
+    }
+    
+    prop = xmlGetNsProp(cur, (const xmlChar *)"NaN", XSLT_NAMESPACE);
+    if (prop != NULL) {
+	if (format->noNumber != NULL) xmlFree(format->noNumber);
+	format->noNumber  = prop;
+    }
+    
+    prop = xmlGetNsProp(cur, (const xmlChar *)"percent", XSLT_NAMESPACE);
+    if (prop != NULL) {
+	if (format->percent != NULL) xmlFree(format->percent);
+	format->percent  = prop;
+    }
+    
+    prop = xmlGetNsProp(cur, (const xmlChar *)"per-mille", XSLT_NAMESPACE);
+    if (prop != NULL) {
+	if (format->permille != NULL) xmlFree(format->permille);
+	format->permille  = prop;
+    }
+    
+    prop = xmlGetNsProp(cur, (const xmlChar *)"zero-digit", XSLT_NAMESPACE);
+    if (prop != NULL) {
+	if (format->zeroDigit != NULL) xmlFree(format->zeroDigit);
+	format->zeroDigit  = prop;
+    }
+    
+    prop = xmlGetNsProp(cur, (const xmlChar *)"digit", XSLT_NAMESPACE);
+    if (prop != NULL) {
+	if (format->digit != NULL) xmlFree(format->digit);
+	format->digit  = prop;
+    }
+    
+    prop = xmlGetNsProp(cur, (const xmlChar *)"pattern-separator", XSLT_NAMESPACE);
+    if (prop != NULL) {
+	if (format->patternSeparator != NULL) xmlFree(format->patternSeparator);
+	format->patternSeparator  = prop;
+    }
+}
+
 /**
  * xsltParseStylesheetPreserveSpace:
  * @style:  the XSLT stylesheet
@@ -445,6 +674,91 @@ xsltParseStylesheetStripSpace(xsltStylesheetPtr style, xmlNodePtr cur) {
     xmlFree(elements);
 }
 
+/**
+ * xsltParseRemoveBlanks:
+ * @style:  the XSLT stylesheet
+ *
+ * Clean-up the stylesheet content from unwanted ignorable blank nodes
+ * and process xslt:text
+ */
+void
+xsltParseRemoveBlanks(xsltStylesheetPtr style) {
+    xmlNodePtr cur, delete;
+
+    /*
+     * This content comes from the stylesheet
+     * For stylesheets, the set of whitespace-preserving
+     * element names consists of just xsl:text.
+     */
+    cur = (xmlNodePtr) style->doc;
+    if (cur == NULL)
+	return;
+    cur = cur->children;
+    delete = NULL;
+    while (cur != NULL) {
+	if (delete != NULL) {
+#ifdef DEBUG_BLANKS
+	    xsltGenericDebug(xsltGenericDebugContext,
+	     "xsltParseRemoveBlanks: removing ignorable blank node\n");
+#endif
+	    xmlUnlinkNode(delete);
+	    xmlFreeNode(delete);
+	    delete = NULL;
+	}
+	if (IS_XSLT_ELEM(cur)) {
+	    if (IS_XSLT_NAME(cur, "text")) {
+		goto skip_children;
+	    }
+	} else if (cur->type == XML_TEXT_NODE) {
+	    if (IS_BLANK_NODE(cur)) {
+		if (xmlNodeGetSpacePreserve(cur) != 1) {
+		    delete = cur;
+		}
+	    }
+	} else if (cur->type != XML_ELEMENT_NODE) {
+	    delete = cur;
+	}
+
+	/*
+	 * Skip to next node
+	 */
+	if (cur->children != NULL) {
+	    if (cur->children->type != XML_ENTITY_DECL) {
+		cur = cur->children;
+		continue;
+	    }
+	}
+skip_children:
+	if (cur->next != NULL) {
+	    cur = cur->next;
+	    continue;
+	}
+	
+	do {
+	    cur = cur->parent;
+	    if (cur == NULL)
+		break;
+	    if (cur == (xmlNodePtr) style->doc) {
+		cur = NULL;
+		break;
+	    }
+	    if (cur->next != NULL) {
+		cur = cur->next;
+		break;
+	    }
+	} while (cur != NULL);
+    }
+    if (delete != NULL) {
+#ifdef DEBUG_PARSING
+	xsltGenericDebug(xsltGenericDebugContext,
+	 "xsltParseRemoveBlanks: removing ignorable blank node\n");
+#endif
+	xmlUnlinkNode(delete);
+	xmlFreeNode(delete);
+	delete = NULL;
+    }
+}
+
 /**
  * xsltParseTemplateContent:
  * @style:  the XSLT stylesheet
@@ -469,9 +783,9 @@ xsltParseTemplateContent(xsltStylesheetPtr style, xsltTemplatePtr ret,
     delete = NULL;
     while (cur != NULL) {
 	if (delete != NULL) {
-#ifdef DEBUG_PARSING
+#ifdef DEBUG_BLANKS
 	    xsltGenericDebug(xsltGenericDebugContext,
-	     "xsltParseStylesheetTemplate: removing ignorable blank node\n");
+	     "xsltParseTemplateContent: removing text\n");
 #endif
 	    xmlUnlinkNode(delete);
 	    xmlFreeNode(delete);
@@ -492,6 +806,10 @@ xsltParseTemplateContent(xsltStylesheetPtr style, xsltTemplatePtr ret,
 				(const xmlChar *)"disable-output-escaping",
 				            XSLT_NAMESPACE);
 			if (prop != NULL) {
+#ifdef DEBUG_PARSING
+			    xsltGenericDebug(xsltGenericDebugContext,
+				 "Disable escaping: %s\n", text->content);
+#endif
 			    if (xmlStrEqual(prop, (const xmlChar *)"yes")) {
 				text->name = xmlStringTextNoenc;
 			    } else if (!xmlStrEqual(prop,
@@ -509,14 +827,6 @@ xsltParseTemplateContent(xsltStylesheetPtr style, xsltTemplatePtr ret,
 		delete = cur;
 		goto skip_children;
 	    }
-	} else if (cur->type == XML_TEXT_NODE) {
-	    if (IS_BLANK_NODE(cur)) {
-		if (xmlNodeGetSpacePreserve(cur) != 1) {
-		    delete = cur;
-		}
-	    }
-	} else if (cur->type != XML_ELEMENT_NODE) {
-	    delete = cur;
 	}
 
 	/*
@@ -551,7 +861,7 @@ skip_children:
     if (delete != NULL) {
 #ifdef DEBUG_PARSING
 	xsltGenericDebug(xsltGenericDebugContext,
-	 "xsltParseStylesheetTemplate: removing ignorable blank node\n");
+	 "xsltParseStylesheetTemplate: removing text\n");
 #endif
 	xmlUnlinkNode(delete);
 	xmlFreeNode(delete);
@@ -563,17 +873,6 @@ skip_children:
      */
     cur = template->children;
     while (cur != NULL) {
-	/*
-	 * Remove Blank nodes found at this level.
-	 */
-	if (IS_BLANK_NODE(cur)) {
-	    xmlNodePtr blank = cur;
-
-            cur = cur->next;
-	    xmlUnlinkNode(blank);
-	    xmlFreeNode(blank);
-	    continue;
-	}
 	if ((IS_XSLT_ELEM(cur)) && (!(IS_XSLT_NAME(cur, "param"))))
 	    break;
 	cur = cur->next;
@@ -583,17 +882,6 @@ skip_children:
      * Browse the remaining of the template
      */
     while (cur != NULL) {
-	/*
-	 * Remove Blank nodes found at this level.
-	 */
-	if (IS_BLANK_NODE(cur)) {
-	    xmlNodePtr blank = cur;
-
-            cur = cur->next;
-	    xmlUnlinkNode(blank);
-	    xmlFreeNode(blank);
-	    continue;
-	}
 	if ((IS_XSLT_ELEM(cur)) && (IS_XSLT_NAME(cur, "param"))) {
 	    xmlNodePtr param = cur;
 
@@ -611,6 +899,98 @@ skip_children:
     ret->content = template->children;
 }
 
+/**
+ * xsltParseStylesheetKey:
+ * @style:  the XSLT stylesheet
+ * @key:  the "key" element
+ *
+ * parse an XSLT stylesheet key definition and register it
+ */
+
+void
+xsltParseStylesheetKey(xsltStylesheetPtr style, xmlNodePtr key) {
+    xmlChar *prop = NULL;
+    xmlChar *use = NULL;
+    xmlChar *match = NULL;
+    xmlChar *name = NULL;
+    xmlChar *nameURI = NULL;
+
+    if (key == NULL)
+	return;
+
+    /*
+     * Get arguments
+     */
+    prop = xmlGetNsProp(key, (const xmlChar *)"name", XSLT_NAMESPACE);
+    if (prop != NULL) {
+	xmlChar *prefix = NULL;
+
+	name = xmlSplitQName2(prop, &prefix);
+	if (name != NULL) {
+	    if (prefix != NULL) {
+		xmlNsPtr ns;
+
+		ns = xmlSearchNs(key->doc, key, prefix);
+		if (ns == NULL) {
+		    xsltGenericError(xsltGenericErrorContext,
+			"no namespace bound to prefix %s\n", prefix);
+		    xmlFree(prefix);
+		    xmlFree(name);
+		    name = prop;
+		    nameURI = NULL;
+		} else {
+		    nameURI = xmlStrdup(ns->href);
+		    xmlFree(prefix);
+		    xmlFree(prop);
+		}
+	    } else {
+		xmlFree(prop);
+		nameURI = NULL;
+	    }
+	} else {
+	    name = prop;
+	    nameURI = NULL;
+	}
+#ifdef DEBUG_PARSING
+	xsltGenericDebug(xsltGenericDebugContext,
+	     "xslt:key: name %s\n", name);
+#endif
+    } else {
+	xsltGenericError(xsltGenericErrorContext,
+	    "xsl:key : error missing name\n");
+	goto error;
+    }
+
+    match = xmlGetNsProp(key, (const xmlChar *)"match", XSLT_NAMESPACE);
+    if (match == NULL) {
+	xsltGenericError(xsltGenericErrorContext,
+	    "xsl:key : error missing match\n");
+	goto error;
+    }
+
+    use = xmlGetNsProp(key, (const xmlChar *)"use", XSLT_NAMESPACE);
+    if (use == NULL) {
+	xsltGenericError(xsltGenericErrorContext,
+	    "xsl:key : error missing use\n");
+	goto error;
+    }
+
+    /*
+     * register the key
+     */
+    xsltAddKey(style, name, nameURI, match, use);
+
+error:
+    if (use != NULL)
+	xmlFree(use);
+    if (match != NULL)
+	xmlFree(match);
+    if (name != NULL)
+	xmlFree(name);
+    if (nameURI != NULL)
+	xmlFree(nameURI);
+}
+
 /**
  * xsltParseStylesheetTemplate:
  * @style:  the XSLT stylesheet
@@ -623,6 +1003,8 @@ void
 xsltParseStylesheetTemplate(xsltStylesheetPtr style, xmlNodePtr template) {
     xsltTemplatePtr ret;
     xmlChar *prop;
+    xmlChar *mode;
+    xmlChar *modeURI;
 
     if (template == NULL)
 	return;
@@ -639,6 +1021,44 @@ xsltParseStylesheetTemplate(xsltStylesheetPtr style, xmlNodePtr template) {
     /*
      * Get arguments
      */
+    prop = xmlGetNsProp(template, (const xmlChar *)"mode", XSLT_NAMESPACE);
+    if (prop != NULL) {
+	xmlChar *prefix = NULL;
+
+	mode = xmlSplitQName2(prop, &prefix);
+	if (mode != NULL) {
+	    if (prefix != NULL) {
+		xmlNsPtr ns;
+
+		ns = xmlSearchNs(template->doc, template, prefix);
+		if (ns == NULL) {
+		    xsltGenericError(xsltGenericErrorContext,
+			"no namespace bound to prefix %s\n", prefix);
+		    xmlFree(prefix);
+		    xmlFree(mode);
+		    mode = prop;
+		    modeURI = NULL;
+		} else {
+		    modeURI = xmlStrdup(ns->href);
+		    xmlFree(prefix);
+		    xmlFree(prop);
+		}
+	    } else {
+		xmlFree(prop);
+		modeURI = NULL;
+	    }
+	} else {
+	    mode = prop;
+	    modeURI = NULL;
+	}
+#ifdef DEBUG_PARSING
+	xsltGenericDebug(xsltGenericDebugContext,
+	     "xslt:template: mode %s\n", mode);
+#endif
+    } else {
+	mode = NULL;
+	modeURI = NULL;
+    }
     prop = xmlGetNsProp(template, (const xmlChar *)"match", XSLT_NAMESPACE);
     if (prop != NULL) {
 	if (ret->match != NULL) xmlFree(ret->match);
@@ -686,7 +1106,12 @@ xsltParseStylesheetTemplate(xsltStylesheetPtr style, xmlNodePtr template) {
      * parse the content and register the pattern
      */
     xsltParseTemplateContent(style, ret, template);
-    xsltAddTemplate(style, ret);
+    xsltAddTemplate(style, ret, mode, modeURI);
+
+    if (mode != NULL)
+	xmlFree(mode);
+    if (modeURI != NULL)
+	xmlFree(modeURI);
 }
 
 /**
@@ -716,7 +1141,7 @@ xsltParseStylesheetTop(xsltStylesheetPtr style, xmlNodePtr top) {
 	if (!xmlStrEqual(prop, (const xmlChar *)"1.0")) {
 	    xsltGenericError(xsltGenericErrorContext,
 		"xsl:version: only 1.0 features are supported\n");
-	    TODO /* set up compatibility when not XSLT 1.0 */
+	     /* TODO set up compatibility when not XSLT 1.0 */
 	}
 	xmlFree(prop);
     }
@@ -738,16 +1163,12 @@ xsltParseStylesheetTop(xsltStylesheetPtr style, xmlNodePtr top) {
 	    continue;
 	}
 	if (IS_XSLT_NAME(cur, "import")) {
-	    TODO /* Handle import */
+	    xsltParseStylesheetImport(style, cur);
 	} else
 	    break;
 	cur = cur->next;
     }
     while (cur != NULL) {
-	if (IS_BLANK_NODE(cur)) {
-            cur = cur->next;
-	    continue;
-	}
 	if (!(IS_XSLT_ELEM(cur))) {
 #ifdef DEBUG_PARSING
 	    xsltGenericDebug(xsltGenericDebugContext,
@@ -761,7 +1182,7 @@ xsltParseStylesheetTop(xsltStylesheetPtr style, xmlNodePtr top) {
 	    xsltGenericError(xsltGenericErrorContext,
 		"xsltParseStylesheetTop: ignoring misplaced import element\n");
         } else if (IS_XSLT_NAME(cur, "include")) {
-	    TODO /* Handle include */
+	    xsltParseStylesheetInclude(style, cur);
         } else if (IS_XSLT_NAME(cur, "strip-space")) {
 	    xsltParseStylesheetStripSpace(style, cur);
         } else if (IS_XSLT_NAME(cur, "preserve-space")) {
@@ -769,11 +1190,11 @@ xsltParseStylesheetTop(xsltStylesheetPtr style, xmlNodePtr top) {
         } else if (IS_XSLT_NAME(cur, "output")) {
 	    xsltParseStylesheetOutput(style, cur);
         } else if (IS_XSLT_NAME(cur, "key")) {
-	    TODO /* Handle key */
+	    xsltParseStylesheetKey(style, cur);
         } else if (IS_XSLT_NAME(cur, "decimal-format")) {
-	    TODO /* Handle decimal-format */
+	    xsltParseStylesheetDecimalFormat(style, cur);
         } else if (IS_XSLT_NAME(cur, "attribute-set")) {
-	    TODO /* Handle attribute-set */
+	    xsltParseStylesheetAttributeSet(style, cur);
         } else if (IS_XSLT_NAME(cur, "variable")) {
 	    xsltParseGlobalVariable(style, cur);
         } else if (IS_XSLT_NAME(cur, "param")) {
@@ -784,7 +1205,7 @@ xsltParseStylesheetTop(xsltStylesheetPtr style, xmlNodePtr top) {
 #endif
 	    xsltParseStylesheetTemplate(style, cur);
         } else if (IS_XSLT_NAME(cur, "namespace-alias")) {
-	    TODO /* Handle namespace-alias */
+	    xsltNamespaceAlias(style, cur);
 	} else {
 	    xsltGenericError(xsltGenericErrorContext,
 		"xsltParseStylesheetTop: ignoring unknown %s element\n",
@@ -799,45 +1220,44 @@ xsltParseStylesheetTop(xsltStylesheetPtr style, xmlNodePtr top) {
 }
 
 /**
- * xsltParseStylesheetDoc:
+ * xsltParseStylesheetProcess:
+ * @ret:  the XSLT stylesheet
  * @doc:  and xmlDoc parsed XML
  *
- * parse an XSLT stylesheet building the associated structures
+ * parse an XSLT stylesheet adding the associated structures
  *
  * Returns a new XSLT stylesheet structure.
  */
 
 xsltStylesheetPtr
-xsltParseStylesheetDoc(xmlDocPtr doc) {
-    xsltStylesheetPtr ret;
+xsltParseStylesheetProcess(xsltStylesheetPtr ret, xmlDocPtr doc) {
     xmlNodePtr cur;
 
     if (doc == NULL)
 	return(NULL);
-
-    ret = xsltNewStylesheet();
     if (ret == NULL)
-	return(NULL);
-
+	return(ret);
+    
     /*
-     * First step, locate the xsl:stylesheet element and the
+     * First steps, remove blank nodes,
+     * locate the xsl:stylesheet element and the
      * namespace declaration.
      */
     cur = xmlDocGetRootElement(doc);
     if (cur == NULL) {
         xsltGenericError(xsltGenericErrorContext,
-		"xsltParseStylesheetDoc : empty stylesheet\n");
+		"xsltParseStylesheetProcess : empty stylesheet\n");
 	xsltFreeStylesheet(ret);
 	return(NULL);
     }
 
-    ret->doc = doc;
+    xsltParseRemoveBlanks(ret);
     if ((IS_XSLT_ELEM(cur)) && 
 	((IS_XSLT_NAME(cur, "stylesheet")) ||
 	 (IS_XSLT_NAME(cur, "transform")))) {
 #ifdef DEBUG_PARSING
 	xsltGenericDebug(xsltGenericDebugContext,
-		"xsltParseStylesheetDoc : found stylesheet\n");
+		"xsltParseStylesheetProcess : found stylesheet\n");
 #endif
 
 	xsltParseStylesheetTop(ret, cur);
@@ -851,20 +1271,20 @@ xsltParseStylesheetDoc(xmlDocPtr doc) {
 	prop = xmlGetNsProp(cur, (const xmlChar *)"version", XSLT_NAMESPACE);
 	if (prop == NULL) {
 	    xsltGenericError(xsltGenericErrorContext,
-		"xsltParseStylesheetDoc : document is not a stylesheet\n");
+		"xsltParseStylesheetProcess : document is not a stylesheet\n");
 	    xsltFreeStylesheet(ret);
 	    return(NULL);
 	}
 
 #ifdef DEBUG_PARSING
         xsltGenericDebug(xsltGenericDebugContext,
-		"xsltParseStylesheetDoc : document is stylesheet\n");
+		"xsltParseStylesheetProcess : document is stylesheet\n");
 #endif
 	
 	if (!xmlStrEqual(prop, (const xmlChar *)"1.0")) {
 	    xsltGenericError(xsltGenericErrorContext,
 		"xsl:version: only 1.0 features are supported\n");
-	    TODO /* set up compatibility when not XSLT 1.0 */
+	     /* TODO set up compatibility when not XSLT 1.0 */
 	}
 	xmlFree(prop);
 
@@ -884,12 +1304,38 @@ xsltParseStylesheetDoc(xmlDocPtr doc) {
 	 * parse the content and register the pattern
 	 */
 	xsltParseTemplateContent(ret, template, (xmlNodePtr) doc);
-	xsltAddTemplate(ret, template);
+	xsltAddTemplate(ret, template, NULL, NULL);
     }
 
     return(ret);
 }
 
+/**
+ * xsltParseStylesheetDoc:
+ * @doc:  and xmlDoc parsed XML
+ *
+ * parse an XSLT stylesheet building the associated structures
+ *
+ * Returns a new XSLT stylesheet structure.
+ */
+
+xsltStylesheetPtr
+xsltParseStylesheetDoc(xmlDocPtr doc) {
+    xsltStylesheetPtr ret;
+
+    if (doc == NULL)
+	return(NULL);
+
+    ret = xsltNewStylesheet();
+    if (ret == NULL)
+	return(NULL);
+    
+    ret->doc = doc;
+    xsltParseStylesheetProcess(ret, doc);
+
+    return(ret);
+}
+
 /**
  * xsltParseStylesheetFile:
  * @filename:  the filename/URL to the stylesheet
