@@ -61,6 +61,27 @@ struct _xsltTemplate {
     xmlNodePtr content;	/* the template replacement value */
 };
 
+/*
+ * Data structure of decimal-format
+ */
+typedef struct _xsltDecimalFormat {
+    struct _xsltDecimalFormat *next; /* chained list */
+    xmlChar *name;
+    /* Used for interpretation of pattern */
+    xmlChar *digit;
+    xmlChar *patternSeparator;
+    /* May appear in result */
+    xmlChar *minusSign;
+    xmlChar *infinity;
+    xmlChar *noNumber; /* Not-a-number */
+    /* Used for interpretation of pattern and may appear in result */
+    xmlChar *decimalPoint;
+    xmlChar *grouping;
+    xmlChar *percent;
+    xmlChar *permille;
+    xmlChar *zeroDigit;
+} xsltDecimalFormat, *xsltDecimalFormatPtr;
+
 /*
  * The in-memory structure corresponding to an XSLT Stylesheet
  * NOTE: most of the content is simply linked from the doc tree
@@ -73,6 +94,7 @@ struct _xsltStylesheet {
      * The stylesheet import relation is kept as a tree
      */
     struct _xsltStylesheet *parent;
+    struct _xsltStylesheet *next;
     struct _xsltStylesheet *imports;
 
     /*
@@ -94,6 +116,7 @@ struct _xsltStylesheet {
     void *templatesHash;	/* hash table or wherever compiled templates
 				   informations are stored */
     void *rootMatch;		/* template based on / */
+    void *keyMatch;		/* template based on key() */
     void *elemMatch;		/* template based on * */
     void *attrMatch;		/* template based on @* */
     void *parentMatch;		/* template based on .. */
@@ -101,6 +124,21 @@ struct _xsltStylesheet {
     void *piMatch;		/* template based on processing-instruction() */
     void *commentMatch;		/* template based on comment() */
     
+    /*
+     * Namespace aliases
+     */
+    xmlHashTablePtr nsAliases;	/* the namespace alias hash tables */
+
+    /*
+     * Attribute sets
+     */
+    xmlHashTablePtr attributeSets;/* the attribute sets hash tables */
+
+    /*
+     * Key definitions
+     */
+    void *keys;				/* key definitions */
+
     /*
      * Output related stuff.
      */
@@ -109,6 +147,9 @@ struct _xsltStylesheet {
     xmlChar *version;		/* version string */
     xmlChar *encoding;		/* encoding string */
     int omitXmlDeclaration;     /* omit-xml-declaration = "yes" | "no" */
+
+    /* Number formatting */
+    xsltDecimalFormatPtr decimalFormat;
     int standalone;             /* standalone = "yes" | "no" */
     xmlChar *doctypePublic;     /* doctype-public string */
     xmlChar *doctypeSystem;     /* doctype-system string */
@@ -120,18 +161,27 @@ struct _xsltStylesheet {
 /*
  * The in-memory structure corresponding to an XSLT Transformation
  */
-typedef enum xsltOutputType {
+typedef enum {
     XSLT_OUTPUT_XML = 0,
     XSLT_OUTPUT_HTML,
     XSLT_OUTPUT_TEXT
 } xsltOutputType;
 
+typedef enum {
+    XSLT_STATE_OK = 0,
+    XSLT_STATE_ERROR,
+    XSLT_STATE_STOPPED
+} xsltTransformState;
+
 typedef struct _xsltTransformContext xsltTransformContext;
 typedef xsltTransformContext *xsltTransformContextPtr;
 struct _xsltTransformContext {
     xsltStylesheetPtr style;		/* the stylesheet used */
     xsltOutputType type;		/* the type of output */
 
+    const xmlChar *mode;		/* the current mode */
+    const xmlChar *modeURI;		/* the current mode URI */
+
     xmlDocPtr doc;			/* the current doc */
     xmlNodePtr node;			/* the current node */
     xmlNodeSetPtr nodeList;		/* the current node list */
@@ -142,17 +192,30 @@ struct _xsltTransformContext {
     xmlXPathContextPtr xpathCtxt;	/* the XPath context */
     void *variablesHash;		/* hash table or wherever variables
 				   	   informations are stored */
+    void *keys;				/* key tables storage */
     xmlDocPtr extraDocs;		/* extra docs parsed by document() */
+    xsltTransformState state;		/* the current state */
 };
 
+#define CHECK_STOPPED if (ctxt->state == XSLT_STATE_STOPPED) return;
+#define CHECK_STOPPEDE if (ctxt->state == XSLT_STATE_STOPPED) goto error;
+#define CHECK_STOPPED0 if (ctxt->state == XSLT_STATE_STOPPED) return(0);
+
 /*
  * Functions associated to the internal types
+xsltDecimalFormatPtr	xsltDecimalFormatGetByName(xsltStylesheetPtr sheet,
+						   xmlChar *name);
  */
 xsltStylesheetPtr	xsltParseStylesheetFile	(const xmlChar* filename);
 void			xsltFreeStylesheet	(xsltStylesheetPtr sheet);
 int			xsltIsBlank		(xmlChar *str);
 void			xsltFreeStackElemList	(xsltStackElemPtr elem);
+xsltDecimalFormatPtr	xsltDecimalFormatGetByName(xsltStylesheetPtr sheet,
+						   xmlChar *name);
 
+xsltStylesheetPtr	xsltParseStylesheetProcess(xsltStylesheetPtr ret,
+	                                         xmlDocPtr doc);
+xsltStylesheetPtr	xsltParseStylesheetDoc	(xmlDocPtr doc);
 #ifdef __cplusplus
 }
 #endif
