@@ -59,6 +59,9 @@ extern void *xsltGenericErrorContext;
 extern xmlGenericErrorFunc xsltGenericDebug;
 extern void *xsltGenericDebugContext;
 
+void		xsltMessage			(xsltTransformContextPtr ctxt,
+						 xmlNodePtr node,
+						 xmlNodePtr inst);
 void		xsltSetGenericErrorFunc		(void *ctx,
 						 xmlGenericErrorFunc handler);
 void		xsltSetGenericDebugFunc		(void *ctx,
@@ -68,6 +71,7 @@ void		xsltSetGenericDebugFunc		(void *ctx,
  * Sorting ... this is definitely a temporary interface !
  */
 
+void		xsltDocumentSortFunction	(xmlNodeSetPtr list);
 void		xsltSortFunction		(xmlNodeSetPtr list,
 						 xmlXPathObjectPtr *results,
 						 int descending,
